/* Some trivial examples of user-defined primitives.
 *
 * Author: Ian.Piumarta@inria.fr
 *
 * Compile this file and link it with the rest of the VM in place of
 * the empty sqUserPrims.c provided.  Then file-in 'userPrimExamples.st'
 * and follow the instructions in the following methods, all of which are
 * found in the 'user primitive examples' protocol on the instance side:
 *
 *	String>>udpPrint
 *	SmallInteger>>udpAdd:
 *	Array>>udpFloatVectorMultiply:
 *	Object>>udpDumpObject
 *	Array>>udpIntegerVectorSort
 *
 * Last edited: Sun Oct 12 03:56:34 1997 by piumarta (Ian Piumarta) on clotho
 */

#include "sqUserPrims.h"

/*   oop implementationFunction(oop recevier [, oop arg1, ..., oop argN])
     { ...
       return resultOop;
     }
 */


/* see: String>>udpPrint */

static oop myPrimitivePrintString(oop receiver)
{
  int length, i;
  if (fetchClassOf(receiver) != String) return primitiveFail();
  length= indexableSize(receiver);
  for (i= 0; i < length; ++i)
    putchar(byteAt(receiver, i));
  putchar('\n');
  return receiver;
}


/* see: SmallInteger>>udpAdd: */

static oop myPrimitiveIntegerAdd(oop receiver, oop argument)
{
  if (!isIntegerObject(receiver) || !isIntegerObject(argument))
    return primitiveFail();
  return integerObjectOf(integerValueOf(receiver) + integerValueOf(argument));
}


/* see: Array>>udpFloatVectorMultiply: */

static oop myPrimitiveFloatVectorMultiply(oop receiver, oop argument)
{
  int i, length;
  oop product;
  /* ensure that receiver and argument are Arrays of the same length */
  if (fetchClassOf(receiver) != Array || fetchClassOf(argument) != Array)
    return primitiveFail();
  length= indexableSize(receiver);
  if (length != indexableSize(argument))
    return primitiveFail();
  /* protect receiver and argument during allocation of product array */
  pushRemappableOop(receiver);
  pushRemappableOop(argument);
  product= new(Array, length);
  /* fill product array with Float objects.  Written using an indexing
     operation since the product array might move during allocation of
     one of its Float elements.  Register it as a root on every store,
     since a GC would leave it in old space but the element in young
     space. */
  for (i= 0; i < length; i++) {
    oop newFloat;
    /* protect product array during allocation of each Float element */
    pushRemappableOop(product);
    newFloat= new(Float, 2);	/* Floats are variableWord of size 2! */
    product= popRemappableOop();
    objectAtPut(product, i, newFloat);
    /* possible store of young object into old object */
    registerRoot(product);
  }
  /* pop the (possibly moved) GC-protected receiver and argument */
  argument= popRemappableOop();
  receiver= popRemappableOop();
  /* perform the vector multiply, failing if any non-Float is
     encountered.  Written to use pointers rather than indexing in
     order to illustrate the mechanism.  Note that the float access
     macros might expand their arguments more than once (this is
     platform-dependent), so side-effects (auto-increment/decrement in
     particular) must be avoided. */
  {
    oop *rp= (oop *)firstIndexableField(receiver);
    oop *ap= (oop *)firstIndexableField(argument);
    oop *pp= (oop *)firstIndexableField(product);
    for (i= 0; i < length; i++) {
      oop obj;
      double r, a;
      if (!isFloatObject(*rp) || !isFloatObject(*ap))
	return primitiveFail();
      fetchFloat(*rp, r);
      fetchFloat(*ap, a);
      r *= a;
      storeFloat(*pp, r);
      ++rp, ++ap, ++pp;
    }
  }
  return product;
}


/* see: Object>>udpDumpObject */

static char *getClassName(oop class)
{
  oop name;
  int i;
  static char result[64];
  if (instVarSize(class) < 6) {
    strcpy(result, "<not a class>");
    return result;
  }
  name= instVarAt(class, 6);	/* 7th inst var is class name (a Symbol) */
  if (!isBytes(name)) {		/* metaclass: 7th inst var is sole instance */
    getClassName(name);		/* the sole instance's name */
    strcat(result, " class");	/* result is static */
    return result;
  }
  for (i= 0; i < indexableSize(name); i++)
    result[i]= byteAt(name, i);
  result[i]= '\0';
  return result;
}

static void printObject(oop object)
{
  if (isIntegerObject(object))
    printf("=%d", integerValueOf(object));
  else
    printf("%s", getClassName(fetchClassOf(object)));
}

static oop myPrimitiveDumpObject(oop receiver)
{
  int nFixed, nIndexed, i;
  if (isIntegerObject(receiver)) {
    printf("=%d\n", integerValueOf(receiver));
    return receiver;
  }
  printObject(receiver);
  printf(":\n");
  nFixed= instVarSize(receiver);
  nIndexed= indexableSize(receiver);
  printf("  %d instance variables\n", nFixed);
  if (nFixed) {
    for (i= 0; i < nFixed; ++i) {
      printf("    %d: ", i+1);
      printObject(instVarAt(receiver, i));
      printf("\n");
    }
  }
  printf("  %d indexable ", nIndexed);
  if (nIndexed) {
    if (isBytes(receiver)) {
      printf("bytes\n", nIndexed);
      for (i= 0; i < nIndexed; ++i)
	printf("    %d: %d\n", i+1, byteAt(receiver, i));
    } else if (isWords(receiver)) {
      printf("words\n", nIndexed);
      for (i= 0; i < nIndexed; ++i)
	printf("    %d: %d\n", i+1, wordAt(receiver, i));
    } else if (isPointers(receiver)) {
      printf("pointers\n", nIndexed);
      for (i= 0; i < nIndexed; ++i) {
	printf("    %d: ", i+1);
	printObject(objectAt(receiver, i));
	printf("\n");
      }
    } else {
      printf("fields of an unknown disposition\n");
    }
  } else
    printf("fields\n");
  return receiver;
}


/* see: Array>>udpIntegerVectorSort */

static int compareInts(const void *a, const void *b)
{
  return integerValueOf(*(oop *)a) - integerValueOf(*(oop *)b);
}

static oop myPrimitiveIntVecSort(oop receiver)
{
  int length;
  if (fetchClassOf(receiver) != Array) return primitiveFail();
  length= indexableSize(receiver);
  qsort(firstIndexableField(receiver), length, sizeof(oop), compareInts);
  return receiver;
}


/* user primitive initialisation routine */

void sqInitializeUserPrimitives(void)
{
  /* declarePrimIndexFunctionArgs(primIndex, implFunction, argCount); ... */
  declarePrimitiveIndexFunctionArgs(1024, myPrimitivePrintString, 0);
  declarePrimitiveIndexFunctionArgs(1025, myPrimitiveIntegerAdd, 1);
  declarePrimitiveIndexFunctionArgs(1026, myPrimitiveFloatVectorMultiply, 1);
  declarePrimitiveIndexFunctionArgs(1027, myPrimitiveDumpObject, 0);
  declarePrimitiveIndexFunctionArgs(1028, myPrimitiveIntVecSort, 0);
}
